import os
import sys

NoResult = None
NoError = None

DB_SCHEMA = "io_fintix"
DB_URL_KEY = "quarkus.datasource.jdbc.url"
DB_USER_KEY = "quarkus.datasource.username"
DB_PWD_KEY = "quarkus.datasource.password"

LOGFILE = "fintix.log"
if sys.platform == "win32":
    USERNAME = os.environ.get("USERNAME", None)
    FINTIX_HOME_DIR = os.path.join(os.environ.get("USERPROFILE", os.path.join("C:\\Users\\", USERNAME)), ".fintix")
    FINTIX_IO_DATA_DIR = os.path.join(FINTIX_HOME_DIR, "data")
    FINTIX_IO_LOG_DIR = os.path.join(FINTIX_HOME_DIR, "logs", "modelcurator")
else:
    FINTIX_IO_DATA_DIR = "/var/lib/fintix"
    FINTIX_IO_LOG_DIR = "/var/log/fintix"

LOCAL_DEPLOY_MODEL_REPO = os.path.join(FINTIX_IO_DATA_DIR, "models", "deploy")
LOCAL_TRAINING_MODEL_REPO = os.path.join(FINTIX_IO_DATA_DIR, "models", "train")

MILLISECOND_PER_UNIT = {
    "s": 1000,
    "m": 60 * 1000,
    "h": 60 * 60 * 1000,
    "d": 24 * 60 * 60 * 1000
}



import os
import sys
import logging

from fintix_modelcurator.callback import FintixCallback
from fintix_modelcurator.utils import get_model_save_path, upload_model
from fintix_modelcurator.config import Config
from fintix_modelcurator.const import *

__all__ = ['FintixCallback', 'get_model_save_path', 'upload_model']


def __init_logging():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)


def __init_config():
    config_file = os.path.join(FINTIX_HOME_DIR, "config", "application.properties")
    Config.getInstance().init(config_file)


__init_logging()
__init_config()



import logging
import subprocess
import importlib
from urllib.parse import urlparse

import psycopg2
from psycopg2 import pool

from fintix_modelcurator.const import *
from fintix_modelcurator.exit_code import *
from fintix_modelcurator.phase import *
from fintix_modelcurator.config import Config
from fintix_modelcurator.error import FintixError
from fintix_modelcurator.utils import *


class ModelRepository:
    INSTANCE = None
    CONN_POOL_MIN = 1
    CONN_POOL_MAX = 5
    MODEL_REPO_TABLE = f"{DB_SCHEMA}.custom_model"

    def __init__(self):
        self.config = Config.getInstance()
        self.db_url, _ = self.config.get(DB_URL_KEY)
        self.db_user, _ = self.config.get(DB_USER_KEY)
        self.db_pwd, _ = self.config.get(DB_PWD_KEY)

        url_parsed = urlparse(self.db_url.replace("jdbc:", ""))
        self.conpool = psycopg2.pool.ThreadedConnectionPool(
            minconn=self.CONN_POOL_MIN,
            maxconn=self.CONN_POOL_MAX,
            user=self.db_user,
            password=self.db_pwd,
            database=url_parsed.path[1:],
            host=url_parsed.hostname,
            port=url_parsed.port
        )

    def __download_model(self, model_name, phase):
        conn = None
        cur = None
        try:
            conn = self.conpool.getconn()
            cur = conn.cursor()
            if conn:
                cur.execute("SELECT model from %s where model_name = '%s'", (ModelRepository.MODEL_REPO_TABLE, model_name))
                blob = cur.fetchone()
                output_dir = LOCAL_TRAINING_MODEL_REPO if phase in [TRAINING_PHASE,
                                                                    EVALUATION_PHASE] else LOCAL_DEPLOY_MODEL_REPO
                output_file = os.path.join(output_dir, model_name + ".py")
                with open(output_file, "wb") as modelzip:
                    modelzip.write(blob[0])
                conn.commit()
        except Exception as e:
            return (
                NoResult,
                FintixError(exception=e, message=f"could not download model {model_name} from model repository",
                            exit_code=MODEL_DOWNLOADING_ERROR, should_exit=True)
            )
        finally:
            if cur is not None:
                try:
                    cur.close()
                except:
                    pass
            if conn is not None:
                try:
                    self.conpool.putconn(conn)
                except:
                    pass
        return NoResult, NoError

    def __convert_model(self, model_name):
        try:
            current_dir = os.getcwd()
            model_name = model_name if not model_name.endswith(".ipynb") else model_name.replace(".ipynb", "")
            model_notebook_path = os.path.join(current_dir, model_name + ".ipynb")
            res = subprocess.getoutput(f"jupyter nbconvert --to script {model_notebook_path}")
            return res, NoError
        except Exception as e:
            return (
                NoResult,
                FintixError(exception=e, message="failed to convert notebook into executable python script",
                            exit_code=None, should_exit=False)
            )

    def upload_model(self, model_name):
        res, err = self.__convert_model(model_name=model_name)
        if err is not NoError:
            handle_error(err)
        else:
            logging.info(res)

        python_script_content = None
        try:
            current_dir = os.getcwd()
            model_name = model_name if not model_name.endswith(".ipynb") else model_name.replace(".ipynb", "")
            python_script_path = os.path.join(current_dir, model_name + ".py")
            with open(python_script_path, "rb") as python_script:
                python_script_content = python_script.read()
            if python_script_content is None:
                raise Exception("failed to read converted python executable")
        except Exception as e:
            return (
                NoResult,
                FintixError(exception=e, message="failed to zip converted python executable",
                            exit_code=None, should_exit=False)
            )

        conn = None
        cur = None
        try:
            conn = self.conpool.getconn()
            cur = conn.cursor()
            cur.execute("INSERT INTO io_fintix.custom_model (model_uid, model_name, model) VALUES(%s, %s, %s)",
                        (model_name, model_name, psycopg2.Binary(python_script_content)))
            conn.commit()
        except Exception as e:
            return (
                NoResult,
                FintixError(exception=e, message="failed to upload converted python to database", exit_code=None,
                            should_exit=False)
            )
        finally:
            if cur is not None:
                try:
                    cur.close()
                except:
                    pass
            if conn is not None:
                try:
                    self.conpool.putconn(conn)
                except:
                    pass

        return NoResult, NoError

    def import_model(self, model_name, phase):
        if model_name is None or phase is None:
            return NoResult, FintixError(exception=None,
                                         message="failed to import model, model or phase could not be done")

        _, err = self.__download_model(model_name, phase)
        if err:
            return NoResult, err
        model_dir = LOCAL_TRAINING_MODEL_REPO if phase in [TRAINING_PHASE,
                                                           EVALUATION_PHASE] else LOCAL_DEPLOY_MODEL_REPO
        try:
            model = importlib.import_module(model_name, model_dir)
        except Exception as e:
            return (
                NoResult,
                FintixError(
                    exception=e, message=f"failed to import model in {model_dir}",
                    exit_code=MODEL_CODE_IMPORT_ERR, should_exit=True)
            )

        return model, NoError

    @classmethod
    def getInstance(cls):
        if cls.INSTANCE is None:
            cls.INSTANCE = ModelRepository()
        return cls.INSTANCE



OK = 0
UNKNOWN = 255

CONFIG_INITIALIZATION_ERROR = 1
INVALID_SETTINGS = 2
MODEL_NOT_DEFINED = 3
PHASE_NOT_DEFINED = 4
INPUT_TYPE_NOT_DEFINED = 5
INVALID_INPUT_TYPE = 6

MODEL_CODE_IMPORT_ERR = 7
CONFIG_NOT_INITIALIZED = 8
INVALID_CONFIG_PATH = 9
MODEL_DOWNLOADING_ERROR = 10
INVALID_PHASE = 11
CALLBACK_IS_NOT_CALLABLE = 12
INPUT_TOPIC_NOT_DEFINED = 13


import os
import time
import logging

from fintix_modelcurator.const import *
from fintix_modelcurator.repository import ModelRepository


def get_currenttime_ms():
    return time.time() // 1_000_000


def handle_error(error, logger=None):
    if logger is None:
        logger = logging

    if error is not None:
        message = "[Unknown message]" if error.getMessage() is None else error.getMessage()
        exception = error.getException()
        exit_code = error.getExitCode()
        shouldExit = error.shouldExit()

        if exception is not None:
            logger.error(message, exception)
        else:
            logger.error(message)

        if shouldExit:
            os._exit(exit_code)


def upload_model(model_name):
    res, err = ModelRepository.getInstance().upload_model(model_name)
    if err is None:
        logging.info(res)
    else:
        handle_error(err)


def get_model_save_path(model_name):
    return os.path.join(LOCAL_DEPLOY_MODEL_REPO, model_name)



from fintix_modelcurator.exit_code import *


class FintixError:
    def __init__(self, exception=None, message=None, exit_code=UNKNOWN, should_exit=False):
        self.exception = exception
        self.message = message
        self.exit_code = exit_code
        self.should_exit = should_exit

    def getException(self):
        return self.exception

    def getMessage(self):
        return self.message

    def getExitCode(self):
        return self.exit_code

    def shouldExit(self):
        return self.should_exit


import logging
from argparse import ArgumentParser

from fintix_modelcurator.const import *
from fintix_modelcurator.exit_code import *
from fintix_modelcurator.trigger_type import *
from fintix_modelcurator.config import Config
from fintix_modelcurator.settings import Settings
from fintix_modelcurator.repository import ModelRepository
from fintix_modelcurator.error import FintixError
from fintix_modelcurator.utils import handle_error
from fintix_modelcurator.cbhandler import DataCallbackHandler


def init_logging():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    return NoResult, NoError


def start():
    setting = Settings.getInstance().get_settings()
    model_name = setting.get('model')
    phase = setting.get('phase')
    data_source = setting.get('data_source')
    input_topic = setting.get('input_topic')
    trigger_type = setting.get('trigger').get('type')
    trigger_interval = setting.get('trigger').get('interval')
    trigger_unit = setting.get('trigger').get('unit')
    stop_after_triggers = setting.get('stop_after')

    if model_name is None:
        error = FintixError(
            exception=None, message="model name was not defined",
            exit_code=MODEL_NOT_DEFINED, should_exit=True)
        return NoResult, error

    if phase is None:
        error = FintixError(
            exception=None, message="phase was not defined",
            exit_code=PHASE_NOT_DEFINED, should_exit=True)
        return NoResult, error

    if trigger_type is None:
        error = FintixError(
            exception=None, message="input type was not defined",
            exit_code=INPUT_TYPE_NOT_DEFINED, should_exit=True)
        return NoResult, error

    if input_topic is None:
        error = FintixError(
            exception=None, message="input topic was not defined",
            exit_code=INPUT_TOPIC_NOT_DEFINED, should_exit=True)
        return NoResult, error

    if trigger_type not in [POINT_INTERVAL, TIME_INTERVAL]:
        error = FintixError(
            exception=None, message="invalid input type",
            exit_code=INVALID_INPUT_TYPE, should_exit=True)
        return NoResult, error

    model, error = ModelRepository.getInstance().import_model(model_name=model_name, phase=phase)
    if error is not NoError:
        return (
            NoResult, error
        )

    callback_handler = DataCallbackHandler.getInstance()
    callback_handler.init(
                        model=model,
                        model_name=model_name,
                        phase=phase,
                        data_source=data_source,
                        input_topic=input_topic,
                        trigger_type=trigger_type,
                        trigger_interval=trigger_interval,
                        trigger_unit=trigger_unit,
                        stop_after_triggers=stop_after_triggers,
                    )

    res, error = callback_handler.start()

    return res, error


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("-c", "--config", help="specified config file", required=True)
    parser.add_argument("-s", "--settings", help="specified settings for data as json string", required=True)
    args = None
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        sys.exit(0)
    else:
        args = parser.parse_args(sys.argv[1:])

    _, error = Config.getInstance().init(args.config)
    handle_error(error)

    _, error = Settings.getInstance().init(args.settings)
    handle_error(error)

    _, error = init_logging()
    handle_error(error)

    _, error = start()
    handle_error(error)

    os._exit(OK)



import json
from fintix_modelcurator.const import *


class Settings:
    INSTANCE = None

    def __init__(self):
        self.settings = None

    def init(self, setting_strings):
        self.settings = json.loads(setting_strings)
        return NoResult, NoError

    def get_settings(self):
        return self.settings

    @classmethod
    def getInstance(cls):
        if cls.INSTANCE is None:
            cls.INSTANCE = Settings()
        return cls.INSTANCE



from configparser import ConfigParser

from fintix_modelcurator.const import *
from fintix_modelcurator.exit_code import *
from fintix_modelcurator.error import FintixError


class Config:
    INSTANCE = None
    CONFIG_DELIMITER = "/"
    SUPPORTED_FILE_EXT = [".properties"]

    def __init__(self):
        self.config = None
        self.initialized = False
        self.config_path = None

    def init(self, config_path):
        self.config_path = config_path
        if not any([self.config_path.endswith(ext) for ext in self.SUPPORTED_FILE_EXT]):
            return (
                NoResult,
                FintixError(
                    exception=Exception(
                        f"only the following config file format are supported: {', '.join(self.SUPPORTED_FILE_EXT)}"
                    ),
                    message="config initialization failed",
                    exit_code=CONFIG_INITIALIZATION_ERROR,
                    should_exit=True
                )
            )

        try:
            with open(self.config_path, 'r') as f:
                content = f.read()
                content = "[DEFAULT]" + os.linesep + content
                self.config = ConfigParser()
                self.config.read_string(content)
        except Exception as e:
            return (
                NoResult,
                FintixError(
                    exception=e,
                    message="config initialization failed",
                    exit_code=CONFIG_INITIALIZATION_ERROR,
                    should_exit=True
                )
            )

        self.initialized = True
        return NoResult, NoError

    def get(self, key_path):
        if self.config is None or not self.initialized:
            return (
                NoResult,
                FintixError(
                    exception=Exception("config have not been initialized"),
                    message="config have not been initialized yet",
                    exit_code=CONFIG_NOT_INITIALIZED,
                    should_exit=False
                )
            )

        if key_path is not None and len(key_path) > 0:
            keys = key_path.strip().split(Config.CONFIG_DELIMITER)
            current_configpath = ""
            current_config = self.config["DEFAULT"]
            for key in keys:
                current_configpath = key if len(current_config) == 0 else Config.CONFIG_DELIMITER.join(
                    (current_configpath, key))
                current_config = current_config.get(key)
                if current_config is None:
                    raise (
                        NoError,
                        FintixError(
                            exception=Exception(f"Could not found config for {key_path} because config already end at {current_configpath}"),
                            message=f"Could not found config for {key_path} because config already end at {current_configpath}",
                            exit_code=INVALID_CONFIG_PATH,
                            should_exit=False

                        )
                    )

            return current_config, NoError
        else:
            return (
                NoResult,
                FintixError(
                    exception=Exception("config key path could not be None or empty"),
                    message="config key path could not be None or empty",
                    exit_code=INVALID_CONFIG_PATH,
                    should_exit=False
                )
            )

    @classmethod
    def getInstance(cls):
        if cls.INSTANCE is None:
            cls.INSTANCE = Config()
        return cls.INSTANCE



import concurrent.futures

from confluent_kafka import Consumer, KafkaError, KafkaException

from fintix_modelcurator.utils import *
from fintix_modelcurator.const import *
from fintix_modelcurator.phase import *
from fintix_modelcurator.trigger_type import *
from fintix_modelcurator.config import Config

_EXECUTOR = concurrent.futures.ThreadPoolExecutor(max_workers=1)


class DataCallbackHandler:
    INSTANCE = None

    def __init__(self):
        self.model = None
        self.phase = None
        self.model_name = None
        self.config = None
        self.kafka_settings = None
        self.consumer = None
        self.producer = None
        self.input_topic = None
        self.trigger_type = None
        self.trigger_interval = None
        self.trigger_unit = None
        self.stop_after_trigger = None

    def init(self,
             model=None,
             model_name=None,
             phase=None,
             data_source=None,
             input_topic=None,
             trigger_type=None,
             trigger_interval=None,
             trigger_unit=None,
             stop_after_triggers=None
            ):
        self.model = model
        self.phase = phase
        self.model_name = model_name
        self.config = Config.getInstance()
        self.kafka_settings = {
            'bootstrap.servers': f'{data_source}',
            'group.id': f'{model_name}-model-handler',
            'client.id': f'{model_name}-model-handler_client',
            'enable.auto.commit': True,
            'session.timeout.ms': 6000,
            'default.topic.config': {
                'auto.offset.reset': 'latest'
            }
        }
        self.input_topic = input_topic
        self.trigger_type = trigger_type
        self.trigger_unit = trigger_unit
        if trigger_type is TIME_INTERVAL:
            self.trigger_interval = trigger_interval * MILLISECOND_PER_UNIT.get(self.trigger_unit)
        self.stop_after_trigger = stop_after_triggers
        self.consumer = Consumer(self.kafka_settings)
        self.consumer.subscribe([self.input_topic])

    def should_trigger(self, data, start):
        if self.trigger_type is POINT_INTERVAL and len(data) > self.trigger_interval:
            return True

        if self.trigger_type is TIME_INTERVAL:
            now = get_currenttime_ms()
            if (now - start) > self.trigger_interval:
                return True

        return False

    def start(self, loop=None):
        data = []
        on_start_cb = None
        on_new_data_cb = None
        on_stop_cb = None
        if self.phase == TRAINING_PHASE:
            on_start_cb = self.model.onTrainingStart
            on_new_data_cb = self.model.onNewTrainingData
            on_stop_cb = self.model.onTrainingStop
        if self.phase == EVALUATION_PHASE:
            on_start_cb = self.model.onEvaluationStart
            on_new_data_cb = self.model.onNewEvaluationData
            on_stop_cb = self.model.onEvaluationStop
        if self.phase == PREDICTION_PHASE:
            on_start_cb = self.model.onPredictionStart
            on_new_data_cb = self.model.onNewPredictionData
            on_stop_cb = self.model.onPredictionStop

        on_start_cb()
        try:
            self.consumer.subscribe([self.input_topic])
            start = get_currenttime_ms()
            trigger_count = 0
            while True:
                msg = loop.run_in_executor(_EXECUTOR, self.consumer.poll, timeout=1.0)
                if msg is None:
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        logging.error(
                            '%% %s [%d] reached end at offset %d\n' % (msg.topic(), msg.partition(), msg.offset()))
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    self.consumer.commit(msg)
                    data.append(msg)
                    if self.should_trigger(data, start):
                        on_new_data_cb(data)
                        data = []
                        trigger_count += 1
                        start = get_currenttime_ms()

                if 0 < self.stop_after_trigger <= trigger_count: # stop_after_trigger < 0 means run forever
                    break
        finally:
            self.consumer.close()
        on_stop_cb()

    @classmethod
    def getInstance(cls):
        if cls.INSTANCE is None:
            cls.INSTANCE = DataCallbackHandler()
        return cls.INSTANC



import json
import logging

from tensorflow.keras.callbacks import Callback

from fintix_modelcurator.config import *
from confluent_kafka import Producer


class FintixCallback(Callback):
    def __init__(self, model_name):
        self.model_name = model_name
        self.config = Config.getInstance()
        self.kafka_settings = {
            'bootstrap.servers': '192.168.2.102:9092',
        }
        self.predict_output_topic = f"{model_name}_model_output"
        self.producer = Producer(self.kafka_settings)
        super(Callback, self).__init__()

    def on_data_delivered(self, err, msg):
        if err is not None:
            logging.error('Message delivery failed: {}'.format(err))
        else:
            logging.error('Message delivered to {} [{}]'.format(msg.topic(), msg.partition()))

    def on_train_end(self, logs=None):
        self.producer.produce(
            f"_fintix_model_monitoring_{self.model_name}_train_end",
            json.dumps(logs).encode('utf-8'),
            self.on_data_delivered)
        self.producer.flush()
        logging.info(f"Stop training; got logs: {logs}")

    def on_train_batch_end(self, batch, logs=None):
        self.producer.produce(
            f"_fintix_model_monitoring_{self.model_name}_train_batch",
            json.dumps(logs).encode('utf-8'),
            self.on_data_delivered)
        self.producer.flush()
        logging.info(f"Training: end of batch {batch}; got logs {logs}")

    def on_epoch_end(self, epoch, logs=None):
        self.producer.produce(
            f"_fintix_model_monitoring_{self.model_name}_train_epoch",
            json.dumps(logs).encode('utf-8'),
            self.on_data_delivered)
        self.producer.flush()
        logging.info(f"End epoch {epoch} of training; got logs {logs}")

    def on_test_end(self, logs=None):
        self.producer.produce(
            f"_fintix_model_monitoring_{self.model_name}_test",
            json.dumps(logs).encode('utf-8'),
            self.on_data_delivered)
        self.producer.flush()
        logging.info(f"Stop testing; got logs {logs}")

    def on_test_batch_end(self, batch, logs=None):
        self.producer.produce(
            f"_fintix_model_monitoring_{self.model_name}_test_batch",
            json.dumps(logs).encode('utf-8'),
            self.on_data_delivered)
        self.producer.flush()
        logging.info(f"Evaluating: end of batch {batch}; got logs {logs}")

    def on_predict_end(self, logs=None):
        self.producer.produce(
            f"_fintix_model_monitoring_{self.model_name}_predict",
            json.dumps(logs).encode('utf-8'),
            self.on_data_delivered)
        self.producer.flush()
        logging.info(f"Stop predicting; got logs {logs}")

    def on_predict_batch_end(self, batch, logs=None):
        for data in logs["outputs"]:
            self.producer.produce(self.predict_output_topic, json.dumps(data).encode('utf-8'), self.on_data_delivered)
            self.producer.flush()
        logging.info(f"Predicting: end of batch {batch}; got logs {logs}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fintix_modelcurator']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'confluent-kafka>=1.6.1,<2.0.0',
 'kafka-python>=2.0.2,<3.0.0',
 'pandas>=1.2.4,<2.0.0',
 'plotly==4.14.3',
 'psycopg2-binary>=2.8.6,<3.0.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'tensorflow-io>=0.17.1,<0.18.0',
 'tensorflow>=2.4.1,<3.0.0',
 'xgboost==1.4.1']

setup_kwargs = {
    'name': 'fintix-modelcurator',
    'version': '0.1.15',
    'description': 'helpers library used for model lab and model training and model deploying',
    'long_description': None,
    'author': 'Trinh Tran',
    'author_email': 'trinhtran2151995@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)
